import Utilities.*;
import Synchronization.*;

class LaplaceGrid extends MyObject implements Runnable {

   private static int M = 4, N = 5, numIterations = 10;
   private static boolean debug = false;
   private static MessagePassing[][] grid = null;
   private static Barrier b = null;

   private int id = -1, jd = -1;
   private Thread thread = null;

   private LaplaceGrid(int id, int jd) {
      super("LaplaceGrid: id=" + id + " jd=" + jd);
      this.id = id;  this.jd = jd;
      thread = new Thread(this);  thread.start();
      if (debug) System.out.println(getName());
   }

   public void join() {
      try { thread.join(); } catch (InterruptedException e) {}
   }

   public void run() {
      double value, got, total, average, correction;
      value = random(10*M*N);
      System.out.println(getName() + " value=" + value);
      for (int iter = 0; iter < numIterations; iter++) {
         if (id < M-2 && jd > 0 && jd < N-1) send(grid[id+1][jd], value);
         if (id > 1   && jd > 0 && jd < N-1) send(grid[id-1][jd], value);
         if (jd < N-2 && id > 0 && id < M-1) send(grid[id][jd+1], value);
         if (jd > 1   && id > 0 && id < M-1) send(grid[id][jd-1], value);
         if (debug) System.out.println(getName() + " messages sent");
         total = 0.0;
         if (id > 0 && jd > 0 && id < M-1 && jd < N-1) {
            for (int i = 0; i < 4; i++) {
               got = receiveDouble(grid[id][jd]);
               total += got;
            }
            average = total/4.0; correction = average - value;
            value += correction;
// optional delay                    nap(1+(int)random(100*M*N));
            if (debug || iter == 0 || iter == numIterations-1)
               System.out.println("iter=" + iter + ", " + getName() 
                  + " new value=" + value + " (correction was "
                  + correction + ")");
         }
         if (debug) System.out.println(getName() + " barrier " + (id+jd*M));
         b.barrier(id + jd*M); // barrier([id][jd]) compute offset 0->M*N-1
      }
   }

   public static void main(String[] args) {

      // parse command line options, if any, to override defaults
      GetOpt go = new GetOpt(args, "Udm:n:i:");
      go.optErr = false;
      String usage = "Usage: -d -m M -n N -i numIterations";
      int ch = -1;
      while ((ch = go.getopt()) != go.optEOF) {
         if      ((char)ch == 'U') {
            System.out.println(usage);  System.exit(0);
         } else if ((char)ch == 'd') {
            debug = true;  // go.optErr = true;
         } else if ((char)ch == 'm') {
            M = go.processArg(go.optArgGet(), M);
            if (M < 3) {
               System.err.println("LaplaceGrid, M < 3");
               System.exit(1);
            }
         } else if ((char)ch == 'n') {
            N = go.processArg(go.optArgGet(), N);
            if (N < 3) {
               System.err.println("LaplaceGrid, N < 3");
               System.exit(1);
            }
         } else if ((char)ch == 'i') {
            numIterations = go.processArg(go.optArgGet(), numIterations);
            if (numIterations < 3) {
               System.err.println("LaplaceGrid, numIterations < 3");
               System.exit(1);
            }
         } else {
            System.err.println(usage);  System.exit(1);
         }
      }
      System.out.println("Laplace Grid, M=" + M + ", N=" + N
         + ", numIterations=" + numIterations);
      
      b = new Barrier(M*N);  // set up the barrier      
      grid = new MessagePassing[M][N];  // set up the communication grid
      for (int m = 0; m < M; m++) for (int n = 0; n < N; n++)
         grid[m][n] = new AsyncMessagePassing();
      
      LaplaceGrid[][] lg = new LaplaceGrid[M][N];  // start the workers
      for (int m = 0; m < M; m++) for (int n = 0; n < N; n++)
         lg[m][n] = new LaplaceGrid(m, n);      
      for (int m = 0; m < M; m++) for (int n = 0; n < N; n++)
         lg[m][n].join();           // wait for them to finish
      b.stop();                     // kill the barrier internal thread
      System.out.println("age()=" + age() + " LaplaceGrid done");
      System.exit(0);
   }
}

/* ............... Example compile and run(s)

D:\Parallel>javac grba.java

D:\Parallel>java LaplaceGrid -i25
Laplace Grid, M=4, N=5, numIterations=25
LaplaceGrid: id=0 jd=0 value=66.3314
LaplaceGrid: id=0 jd=1 value=53.9016
LaplaceGrid: id=0 jd=2 value=83.7027
LaplaceGrid: id=0 jd=3 value=140.071
LaplaceGrid: id=0 jd=4 value=12.473
LaplaceGrid: id=1 jd=0 value=118.412
LaplaceGrid: id=1 jd=1 value=18.9147
LaplaceGrid: id=1 jd=2 value=130.739
LaplaceGrid: id=1 jd=3 value=47.631
LaplaceGrid: id=1 jd=4 value=150.531
LaplaceGrid: id=2 jd=0 value=7.37112
LaplaceGrid: id=2 jd=1 value=190.725
iter=0, LaplaceGrid: id=1 jd=1 new value=123.445 (correction was 104.53)
LaplaceGrid: id=2 jd=3 value=59.6014
LaplaceGrid: id=2 jd=2 value=25.5235
LaplaceGrid: id=2 jd=4 value=149.396
LaplaceGrid: id=3 jd=0 value=62.7077
LaplaceGrid: id=3 jd=1 value=171.039
iter=0, LaplaceGrid: id=1 jd=3 new value=120.236 (correction was 72.6047)
iter=0, LaplaceGrid: id=1 jd=2 new value=43.943 (correction was -86.7965)
iter=0, LaplaceGrid: id=2 jd=1 new value=55.7122 (correction was -135.013)
LaplaceGrid: id=3 jd=2 value=186.103
LaplaceGrid: id=3 jd=3 value=3.89405
LaplaceGrid: id=3 jd=4 value=38.3616
iter=0, LaplaceGrid: id=2 jd=2 new value=141.792 (correction was 116.269)
iter=0, LaplaceGrid: id=2 jd=3 new value=56.6112 (correction was -2.99012)
iter=24, LaplaceGrid: id=1 jd=1 new value=94.3802 (correction was 0.000466217)
iter=24, LaplaceGrid: id=2 jd=1 new value=98.9104 (correction was -0.000355913)
iter=24, LaplaceGrid: id=2 jd=2 new value=122.852 (correction was 0.00065933)
iter=24, LaplaceGrid: id=1 jd=3 new value=124.249 (correction was 0.000466217)
iter=24, LaplaceGrid: id=2 jd=3 new value=100.098 (correction was -0.000355913)
iter=24, LaplaceGrid: id=1 jd=2 new value=106.296 (correction was -0.000503337)
age()=880 LaplaceGrid done
                                            ... end of example run(s)  */
