package Synchronization;

import java.net.*;
import Utilities.*;

/*
 * Extended Rendezvous: a two-way flow of information, much like a
 * remote procedute call made by a client thread to a server thread.
 * A simple rendezvous is synchronous message passing, a one-way flow
 * of information while sender and receiver are blocked.  In the extended
 * kind the client thread sends an object to the server that represents a
 * request and blocks waiting for the reply; the server blocks waiting
 * for a request.  The server gets the request, computes the reply,
 * and sends it to the client, unblocking it.  The server may
 * spawn off a thread to handle the request while the server gets
 * additional requests.  Another paradigm represented by this class is
 * master/worker.  The master waits for requests from the worker threads
 * that they are able to do another piece of work.  The master sends a
 * work command to the worker, who sends the result back later in another
 * rendezvous (using the same ExtendedRendezvous object or calling the
 * EstablishRendezvous object again).  In the same JVM, synchronous message
 * is used because there is no need for buffered sends.  Between JVMs
 * that might be on different physical machines, sockets are used.  The
 * other possibilities (sending raw data types through a pipe or socket)
 * are exercises for the reader.
 */

public class ExtendedRendezvous extends MyObject implements Rendezvous {

   private MessagePassing in = null;
   private MessagePassing out = null;

   public ExtendedRendezvous() {
      super("ExtendedRendezvous");
/*
 * We used synchronous message passing rather than asynchronous because
 * buffering is unnecessary and would waste space; requester blocks on
 * the send in synchronous case and on the receive in asynchronous case.
 */
      this.in = new SyncMessagePassing();
      this.out = new SyncMessagePassing();
   }

   public ExtendedRendezvous(Socket socket) {
      super("ExtendedRendezvous");
      this.in = new ObjPipedMessagePassing(socket);
      this.out = this.in;
   }

/*
 * We do not need synchronized methods here because synchronization is
 * handled inside the send and receive methods.
 */
   public Object clientMakeRequestAwaitReply(Object m) {
      send(in, m);
      return receive(out);
   }

   public Object serverGetRequest() {
      return receive(in);
   }

   public void serverMakeReply(Object m) {
      send(out, m);
   }

   public void close() {
      try {
         in.close();
         out.close();
      } catch (NotImplementedMethodException e) {}
   }

   public String toString() { return getName(); }
}
