package Synchronization;

import java.net.*;
import java.io.*;

// like a bounded buffer with multiple producers and consumers
public final class PipedMessagePassing extends MessagePassingRoot {

   private PipedOutputStream outPipe = null;
   private PipedInputStream inPipe = null;
   private DataOutputStream outData = null;
   private DataInputStream inData = null;
   private Socket socket = null;
   private final Object sending = new Object();
   private final Object receiving = new Object();

   // construct a pipe through which the base data types int and double
   // can be passed in raw or binary format within the same JVM
   public PipedMessagePassing() {
      super("PipedMessagePassing");
      outPipe = new PipedOutputStream();
      try {
         inPipe = new PipedInputStream(outPipe);
      } catch (IOException e) {
         System.err.println("PipedMessagePassing:" + e);
         throw new MessagePassingException();
      }
      outData = new DataOutputStream(outPipe);
      inData = new DataInputStream(inPipe);
   }

   // construct a socket through which the base data types int and double
   // can be passed in raw or binary format between JVMs
   public PipedMessagePassing(Socket socket) {
      super("PipedMessagePassing");
      if (socket == null) {
         System.err.println("null socket passed to PipedMessagePassing");
         throw new NullPointerException();
      }
      this.socket = socket;
      try {
         outData = new DataOutputStream(socket.getOutputStream());
         inData = new DataInputStream(socket.getInputStream());
      } catch (IOException e) {
         System.err.println("PipedMessagePassing:" + e);
         throw new MessagePassingException();
      }
   }

   public String toString() { return "PipedMessagePassing"; }

/*
 * Pipes are supposed to already be thread-safe (page 202 Arnold & Gosling,
 * "The Java Programming Language").  And you can't synchronize these
 * methods anyway since that would lead to deadlock.  But maybe they meant
 * thread-safe with respect to a single writer and a single reader of the
 * pipe.  Maybe they didn't mean multiple readers and writers.  So maybe
 * the bodies of the reading and writing methods need to be synchronized
 * on two different objects.
 *
 * Doing that caused a delay problem: if one or more consumers were blocked
 * on an empty pipe, then they stayed blocked until several producers
 * FILLED the pipe, not just one producer unblocking one consumer.
 *
 * But I have had Parallel\qumw.java (Nqueens) fail every once in a while
 * randomly with 1, 2, and 3 being read as 256, 512, and 768, etc.
 */

   public final void send(int m) {
      synchronized (sending) {
         try {
            outData.writeInt(m);
         } catch (IOException e) {
            System.err.println("PipedMessagePassing.send:" + e);
            throw new MessagePassingException();
         }
      }
   }

   public final void send(double m) {
      synchronized (sending) {
         try {
            outData.writeDouble(m);
         } catch (IOException e) {
            System.err.println("PipedMessagePassing.send:" + e);
            throw new MessagePassingException();
         }
      }
   }

   public final void send(Object m) throws NotImplementedMethodException {
      throw new NotImplementedMethodException();
   }

   public final Object receive() throws NotImplementedMethodException {
      throw new NotImplementedMethodException();
      // return null;  // compiler says "Statement not reached."
   }

   public final int receiveInt() {
      int value = 0;
      synchronized (receiving) {
         try {
            value = inData.readInt();
         } catch (IOException e) {
            System.err.println("PipedMessagePassing.receive:" + e);
            throw new MessagePassingException();
         }
      }
      return value;
   }

   public final double receiveDouble() {
      double value = 0.0;
      synchronized (receiving) {
         try {
            value = inData.readDouble();
         } catch (IOException e) {
            System.err.println("PipedMessagePassing.receive:" + e);
            throw new MessagePassingException();
         }
      }
      return value;
   }

   public void close() {
      try {
         if (inData != null) inData.close();
         if (outData != null) outData.close();
         if (inPipe != null) inPipe.close();
         if (outPipe != null) outPipe.close();
         if (socket != null) socket.close();
      } catch (IOException e) {
         System.err.println(getName() + ", close: " + e);
      }
   }
}

/*************************************************************************
D:\Messaging>javac bagt.java

D:\Messaging>java ProducersConsumers -P2 -C5 -p2 -c2
ProducersConsumers, numProducers=2, numConsumers=5, pNap=2, cNap=2
PIPED message passing is being used.
age=110, PRODUCER0 napping for 1010 ms
age=110, PRODUCER1 napping for 975 ms
age=160, Consumer0 napping for 711 ms
age=160, Consumer1 napping for 1184 ms
age=160, Consumer2 napping for 805 ms
All threads started
age=220, Consumer4 napping for 120 ms
age=220, Consumer3 napping for 954 ms
age=330, Consumer4 wants to consume
age=870, Consumer0 wants to consume
age=980, Consumer2 wants to consume
age=1090, PRODUCER1 produced item 0.363669
age=1090, PRODUCER1 deposited item 0.363669
age=1090, PRODUCER1 napping for 87 ms
age=1090, PRODUCER0 produced item 0.266112
age=1090, PRODUCER0 deposited item 0.266112
age=1090, PRODUCER0 napping for 1686 ms
age=1150, Consumer3 wants to consume
age=1200, PRODUCER1 produced item 0.704872
age=1200, PRODUCER1 deposited item 0.704872
age=1200, PRODUCER1 napping for 673 ms
age=1310, Consumer4 fetched item 0.363669              WHY THE DELAY???
age=1310, Consumer4 napping for 898 ms
age=1310, Consumer0 fetched item 0.266112
age=1310, Consumer0 napping for 1433 ms
age=1310, Consumer2 fetched item 0.704872
age=1310, Consumer2 napping for 1170 ms
age=1310, Consumer1 wants to consume
age=1860, PRODUCER1 produced item 0.990244
age=1860, PRODUCER1 deposited item 0.990244
age=1860, PRODUCER1 napping for 786 ms
age=2250, Consumer4 wants to consume
age=2360, Consumer3 fetched item 0.990244
age=2360, Consumer3 napping for 1412 ms
age=2520, Consumer2 wants to consume
age=2690, PRODUCER1 produced item 0.565785
age=2690, PRODUCER1 deposited item 0.565785
age=2690, PRODUCER1 napping for 427 ms
age=2740, Consumer0 wants to consume
age=2800, PRODUCER0 produced item 0.716008
age=2800, PRODUCER0 deposited item 0.716008
age=2800, PRODUCER0 napping for 1585 ms
age=3130, PRODUCER1 produced item 0.581265
age=3130, PRODUCER1 deposited item 0.581265
age=3130, PRODUCER1 napping for 419 ms
age=3350, Consumer1 fetched item 0.565785
age=3350, Consumer1 napping for 25 ms
age=3350, Consumer4 fetched item 0.716008
age=3350, Consumer4 napping for 893 ms
age=3350, Consumer2 fetched item 0.581265
age=3350, Consumer2 napping for 345 ms
age=3400, Consumer1 wants to consume
age=3510, PRODUCER1 produced item 0.936504
age=3510, PRODUCER1 deposited item 0.936504
age=3510, PRODUCER1 napping for 170 ms
age=3730, PRODUCER1 produced item 0.0608161
age=3730, PRODUCER1 deposited item 0.0608161
age=3730, PRODUCER1 napping for 1563 ms
age=3730, Consumer2 wants to consume
age=3730, Consumer3 wants to consume
age=4280, Consumer4 wants to consume
age=4390, Consumer0 fetched item 0.936504
age=4390, Consumer0 napping for 363 ms
age=4390, Consumer1 fetched item 0.0608161
age=4390, Consumer1 napping for 1059 ms
age=4390, PRODUCER0 produced item 0.775623
age=4390, PRODUCER0 deposited item 0.775623
age=4390, PRODUCER0 napping for 1088 ms
age=4720, Consumer0 wants to consume
         .....
age()=20210, time to stop the threads and exit

**************************************************************************/

/*************************************************************************

Maybe the pipe is not thread-safe wrt to multiple readers and writers.
So I put the synchronization on the sending and receiving objects back in.

D:\Parallel>java qumw.java

D:\Parallel>java Nqueens -n8
Nqueens: N=8 NCPU=4
age()=160, Worker1 counting solutions with column 1 queen in row 256
age()=160, Worker2 counting solutions with column 1 queen in row 512
age()=220, Worker3 counting solutions with column 1 queen in row 768
age()=220, Worker0 counting solutions with column 1 queen in row 4
age()=440 Solutions with column 1 queen in row 256 = 312
age()=710 Solutions with column 1 queen in row 512 = 312
age()=930, Worker2 counting solutions with column 1 queen in row 6
age()=880 Solutions with column 1 queen in row 4 = 18
age()=880 Solutions with column 1 queen in row 768 = 312
age()=880, Worker1 counting solutions with column 1 queen in row 5
age()=990, Worker0 counting solutions with column 1 queen in row 7
age()=930 Solutions with column 1 queen in row 6 = 16
age()=990, Worker3 counting solutions with column 1 queen in row 8
age()=1040 Solutions with column 1 queen in row 5 = 18
age()=1100 Solutions with column 1 queen in row 7 = 8
age()=1150 Solutions with column 1 queen in row 8 = 4
age()=1150, Nqueens: numSolutions=1000

************************************************************************/
